<!doctype html>
<html>
  <head>
    <title>OPAQUE Auth Demo</title>
    <meta http-equiv="Content-Security-Policy" content="default-src *; style-src 'self' 'unsafe-inline'; script-src * 'unsafe-inline' 'unsafe-eval'">
  </head>
  <body>
    <h1>OPAQUE Auth Demo</h1>
    Username: <input type="string" id="id" /><br />
    Password: <input type="password" id="pw" /><br />
    <button name="register">Register</button>
    <button name="authenticate">Authenticate</button>
    <pre></pre>
    <script async src="/static/index.js"></script>
  </body>
</html>
